from collections import defaultdict
from datetime import datetime
from itertools import cycle as i_cycle
import os
from pathlib import Path
from typing import (
        Any, Callable, Iterable, Mapping, Optional,
        Sequence, Set, Tuple, Type, Union,
)


class iterdict(defaultdict):
    def __init__(self, iterable: Iterable, /, cycle=False):
        if cycle:
            iterator = iter(iterable)
        else:
            iterator = i_cycle(iterable)
        self._iterator = iterator
        super().__init__(lambda: next(self._iterator))


class EnvironPath(Path):
    _flavour = type(Path())._flavour  # type: ignore # Necessary to avoid TypeError during instantiation
    _environ_var = None
    _initialized = False

    def __new__(cls, environ_var):
        instance = super().__new__(cls)
        instance._initialized = False
        instance._environ_var = environ_var
        return instance

    @staticmethod
    def _get_env_directory(environ_var) -> Path:
        path = os.environ.get(environ_var)
        if path is None:
            raise EnvironmentError(
                f"Please, set environment variable '{environ_var}'."
            )
        path = Path(path)
        if not path.exists():
            raise FileNotFoundError(
                f"Environment variable {environ_var} set to '{path}', "
                "but directory does not exist."
            )
        return path

    @classmethod
    def _from_parsed_parts(cls, drv, root, parts):
        return Path()._from_parsed_parts(drv, root, parts) # type: ignore

    def _initialize(self):
        if not self._initialized:
            # Adapted from Path._from_parts
            drv, root, parts = self._parse_args(
                [self._get_env_directory(self._environ_var)]
            )
            self._drv = drv
            self._root = root
            self._parts = parts
            self._initialized = True

    def __getattribute__(self, name: str, /) -> Any:
        _require_initialization = ["_make_child"] + \
            [method for method in dir(Path) if not method.startswith("_")]
        if name in _require_initialization and not self._initialized:
            self._initialize()
        return super().__getattribute__(name)

    def __getattr__(self, name: str, /) -> Any:
        if not self._initialized:
            print(name)
            self._initialize()
        super().__getattribute__(name)

    def __repr__(self) -> str:
        return f"<EnvironPath({self._environ_var}, resolved={super().__str__()})>"


def _add_date(datetimes, date):
    if date == "0":
        datetimes.append(datetime(1900, 1, 1))
    elif date == "E":
        datetimes.append(datetime(datetime.now().year + 1, 1, 1))
    else:
        datetimes.append(datetime.fromisoformat(date))

def add_date_parsing_args(parser):
    parser.add_argument(
        "-d",
        "--dates",
        action="store",
        type=str,
        nargs="*",
        metavar="DATES",
    )
    parser.add_argument(
        "-i",
        "--interval",
        action="store_true",
        help="Wether the passed dates are an interval. Then, they must be two and in isoformat.",
    )

def parse_date_interval(dates: Sequence[str]) -> Tuple[datetime, datetime]:
    if len(dates) != 2:
        raise ValueError(
            "When passing an interval, there should only be two dates (min and max)."
        )
    datetimes = []
    try:
        for date in dates:
            _add_date(datetimes, date)
    except ValueError as exc:
        raise ValueError(
            "When passing an interval, dates should be passed in isoformat."
        ) from exc
    return tuple(datetimes)

def process_dct(
    dct: Mapping[Any, Any],
    prefix: str = "",
    valid_types: Optional[Sequence[Type]] = None,
    force_fun: Callable =str,
    list_wrap: bool = False,
    *,
    type_changed_keys: Optional[Sequence] = None,
):
    assert isinstance(dct, dict)
    processed_dict = {}
    for key, value in dct.items():
        processed_key = f"{prefix}{key}"
        if valid_types and force_fun and not isinstance(value, valid_types):
            processed_value = force_fun(value)
            if isinstance(type_changed_keys, set):
                type_changed_keys.add(processed_key)
        else:
            processed_value = value
        if list_wrap:
            processed_value = [processed_value]
        processed_dict[processed_key] = processed_value
    return processed_dict

